from .util import PathResolvable, Snapshot
from aridity import NoSuchPathException
from aridity.config import ConfigCtrl
from aridity.model import Function, Text
from aridity.scope import slashfunction
from lagoon import git
from lagoon.program import ONELINE
import re, subprocess, yaml

singledigit = re.compile('[0-9]')
zeroormorespaces = re.compile(' *')
linefeed = '\n'
toplevelres = PathResolvable('toplevel')

def blockliteral(scope, textresolvable):
    contextindent = scope.resolved('indent').cat()
    text = yaml.dump(textresolvable.resolve(scope).cat(), default_style = '|')
    header, *lines = text.splitlines() # For template interpolation convenience we discard the (insignificant) trailing newline.
    if not lines:
        return Text(header)
    if '...' == lines[-1]:
        lines.pop() # XXX: Could this result in no remaining lines?
    indentunit = scope.resolved('indentunit').cat()
    m = singledigit.search(header)
    if m is None:
        pyyamlindent = len(zeroormorespaces.match(lines[0]).group())
    else:
        pyyamlindent = int(m.group())
        header = f"{header[:m.start()]}{len(zeroormorespaces.fullmatch(indentunit).group())}{header[m.end():]}"
    return Text(f"""{header}\n{linefeed.join(f"{contextindent}{indentunit}{line[pyyamlindent:]}" for line in lines)}""")

def rootpath(scope, *resolvables):
    return slashfunction(scope, toplevelres, *resolvables)

def _toplevel(anydir):
    try:
        return Text(git.rev_parse.__show_toplevel[ONELINE](cwd = anydir))
    except subprocess.CalledProcessError:
        raise NoSuchPathException('Git property: toplevel')

def createparent(soakroot):
    parent = ConfigCtrl().node
    s = (-parent).scope()
    s['|',] = Function(blockliteral)
    s['//',] = Function(rootpath)
    s['toplevel',] = Snapshot(lambda: _toplevel(soakroot))
    (-parent).execute('data = $processtemplate$(from)') # XXX: Too easy to accidentally override?
    parent.indentunit = 4 * ' '
    return parent



'Process aridity templates as per all soak.arid configs in directory tree.'
from . import cpuexecutor
from .context import createparent
from .multifork import Tasks
from .terminal import getterminal, Style
from argparse import ArgumentParser
from diapyr.util import invokeall
from functools import partial
from lagoon import diff
from lagoon.util import atomic
from pathlib import Path
import logging, os, sys

log = logging.getLogger(__name__)

class SoakConfig:

    soakkey = 'soak'

    def __init__(self, parent, configpath):
        self.node = (-parent).childctrl().node
        self.node.cwd = str(configpath.parent.resolve())
        (-self.node).printf(". %s", configpath.name)
        self.reltargets = [Path(rt) for rt, _ in (-getattr(self.node, self.soakkey)).scope().resolvables.items()]
        self.dirpath = configpath.parent

    def process(self, reltarget):
        with atomic(self.dirpath / reltarget) as partpath:
            (-self.node).scope().resolved(self.soakkey, str(reltarget), 'data').writeout(partpath)

    def origtext(self, reltarget):
        return getattr(getattr(self.node, self.soakkey), str(reltarget)).diff

    def diff(self, origtext, reltarget):
        diff._us[print]('--color=always', '-', self.dirpath / reltarget, input = origtext, check = False)

def main():
    logging.basicConfig(format = "[%(levelname)s] %(message)s", level = logging.DEBUG)
    parser = ArgumentParser()
    parser.add_argument('-n', action = 'store_true')
    parser.add_argument('-d', action = 'store_true')
    parser.add_argument('-v', action = 'store_true', help = 'show debug logging')
    config = parser.parse_args()
    if not config.v:
        logging.getLogger().setLevel(logging.INFO)
    soakroot = Path('.')
    parent = createparent(soakroot)
    soakconfigs = [SoakConfig(parent, p) for p in soakroot.rglob('soak.arid')]
    if not config.n:
        terminal = getterminal()
        tasks = Tasks()
        for soakconfig in soakconfigs:
            for reltarget in soakconfig.reltargets:
                task = partial(soakconfig.process, reltarget)
                task.index = len(tasks)
                task.target = soakconfig.dirpath / reltarget
                terminal.head(task.index, task.target, Style.pending)
                tasks.append(task)
        tasks.started = lambda task: terminal.head(task.index, task.target, Style.running)
        tasks.stdout = lambda task, line: terminal.log(task.index, sys.stdout, line)
        tasks.stderr = lambda task, line: terminal.log(task.index, sys.stderr, line)
        tasks.stopped = lambda task, code: terminal.head(task.index, task.target, Style.abrupt if code else Style.normal)
        tasks.drain(os.cpu_count())
    if config.d:
        with cpuexecutor() as executor:
            diffs = []
            for soakconfig in soakconfigs:
                for reltarget in soakconfig.reltargets:
                    diffs.append(partial(lambda soakconfig, origtextfuture, reltarget: soakconfig.diff(origtextfuture.result(), reltarget),
                            soakconfig, executor.submit(soakconfig.origtext, reltarget), reltarget))
            invokeall(diffs)

if '__main__' == __name__:
    main()



from diapyr.util import singleton
from itertools import islice
from lagoon import tput
from subprocess import CalledProcessError
import sys

class Style:

    pending = object()
    running = object()
    normal = object()
    abrupt = object()

class Terminal:

    class Section:

        height = 0

    def __init__(self, width):
        self.sections = []
        self.width = width

    def _common(self, index, tonewh):
        dy = sum(s.height for s in islice(self.sections, index + 1, None))
        section = self.sections[index]
        oldh = section.height
        section.height = newh = tonewh(oldh)
        if dy:
            tput.cuu(dy, stdout = sys.stderr)
        if newh > oldh:
            tput.il(newh - oldh, stdout = sys.stderr)
        return dy, oldh, newh

    def head(self, index, obj, style):
        for _ in range(len(self.sections), index + 1):
            self.sections.append(self.Section())
        dy, oldh, newh = self._common(index, lambda h: max(1, h))
        if oldh:
            tput.cuu(oldh, stdout = sys.stderr)
        if Style.pending == style:
            tput.setaf(0, stdout = sys.stderr)
        elif Style.running == style:
            tput.rev(stdout = sys.stderr)
        elif Style.abrupt == style:
            tput.setab(1, stdout = sys.stderr)
            tput.setaf(7, stdout = sys.stderr)
        sys.stderr.write(f"[{obj}]{tput.sgr0()}\n")
        sys.stderr.write('\n' * (newh - 1 + dy))

    def log(self, index, stream, line):
        dy, oldh, newh = self._common(index, lambda h: h + 1)
        noeol, = line.splitlines()
        eol = line[len(noeol):]
        if noeol:
            chunks = [noeol[i:i + self.width] for i in range(0, len(noeol), self.width)]
            stream.write(chunks[0])
            for c in islice(chunks, 1, None):
                stream.flush()
                tput.hpa(0, stdout = sys.stderr)
                sys.stderr.flush()
                stream.write(c)
        if eol:
            stream.write(eol)
        else:
            stream.flush()
        sys.stderr.write('\n' * ((not eol) + dy))

@singleton
class LogFile:

    words = {
        Style.running: 'Damp',
        Style.normal: 'Soaked',
        Style.abrupt: 'Failed',
    }

    def head(self, index, obj, style):
        try:
            word = self.words[style]
        except KeyError:
            return
        print(f"{word}:", obj, file = sys.stderr)

    def log(self, index, stream, line):
        stream.write(line)

def getterminal():
    try:
        width = int(tput.cols())
    except CalledProcessError:
        return LogFile
    return Terminal(width)



from aridity.model import Resolvable
from threading import Lock

class Snapshot(Resolvable):

    class Result:

        def __init__(self, eornone, obj):
            self.eornone = eornone
            self.obj = obj

        def get(self):
            if self.eornone is not None:
                raise self.eornone
            return self.obj

    def __init__(self, factory):
        self.lock = Lock()
        self.factory = factory

    def _loadresult(self):
        with self.lock:
            try:
                self.result
            except AttributeError:
                f = self.factory
                try:
                    obj = f()
                    eornone = None
                except Exception as e:
                    obj = None
                    eornone = e
                self.result = self.Result(eornone, obj)

    def resolve(self, scope):
        try:
            r = self.result
        except AttributeError:
            self._loadresult()
            r = self.result
        return r.get()

class PathResolvable(Resolvable):

    def __init__(self, *path):
        self.path = path

    def resolve(self, scope):
        return scope.resolved(*self.path)



from base64 import b64decode, b64encode
from diapyr.util import invokeall
from select import select
from tblib import Traceback
import os, pickle, sys

class GoodResult:

    def __init__(self, value):
        self.value = value

    def get(self):
        return self.value

class BadResult:

    @classmethod
    def _of(cls, *args):
        return cls(*args)

    def __init__(self, exception):
        self.exception = exception

    def __getstate__(self):
        c = self.exception.__context__
        return dict(
            context = None if c is None else self._of(c),
            exception = self.exception,
            tb = Traceback(self.exception.__traceback__),
        )

    def __setstate__(self, state):
        self.exception = e = state['exception']
        c = state['context']
        e.__context__ = None if c is None else c.exception
        e.__traceback__ = state['tb'].as_traceback()

    def get(self):
        raise self.exception

class Job:

    ttl = 3

    def __init__(self, index, task):
        self.index = index
        self.task = task

    def start(self):
        r1, w1 = os.pipe()
        r2, w2 = os.pipe()
        rx, wx = os.pipe()
        pid = os.fork()
        if pid:
            os.close(w1)
            os.close(w2)
            os.close(wx)
            self.pid = pid
            return map(os.fdopen, [r1, r2, rx])
        os.close(r1)
        os.close(r2)
        os.close(rx)
        os.dup2(w1, 1)
        os.close(w1)
        os.dup2(w2, 2)
        os.close(w2)
        try:
            obj = GoodResult(self.task())
            code = 0
        except BaseException as e:
            obj = BadResult(e)
            code = 1
        os.write(wx, b64encode(pickle.dumps([self.index, obj])))
        sys.exit(code)

    def decr(self):
        self.ttl = ttl = self.ttl - 1
        if not ttl:
            s = os.waitpid(self.pid, 0)[1]
            return -os.WTERMSIG(s) if os.WIFSIGNALED(s) else os.WEXITSTATUS(s)

class Tasks(list):

    def drain(self, limit):
        def report(task, line):
            index, obj = pickle.loads(b64decode(line))
            results[index] = obj.get
        streams = {}
        running = 0
        results = [None] * len(self)
        while self or streams:
            while self and running < limit:
                job = Job(len(results) - len(self), self.pop(0))
                r1, r2, rx = job.start()
                streams[r1] = job, self.stdout
                streams[r2] = job, self.stderr
                streams[rx] = job, report
                running += 1
                self.started(job.task)
            for r in select(streams, [], [])[0]:
                line = r.readline()
                if line:
                    job, callback = streams[r]
                    callback(job.task, line)
                else:
                    job = streams.pop(r)[0]
                    r.close()
                    code = job.decr()
                    if code is not None:
                        running -= 1
                        self.stopped(job.task, code)
        return invokeall(results)

    def started(self, task):
        pass

    def stdout(self, task, line):
        pass

    def stderr(self, task, line):
        pass

    def stopped(self, task, code):
        pass
